@@ -7,6 +7,13 @@ import dns.resolver
 import dns.exception
 import idna  # implements IDNA 2008; Python's codec is only IDNA 2003
 
+# Default values for keyword arguments.
+
+ALLOW_SMTPUTF8 = True
+CHECK_DELIVERABILITY = True
+TEST_ENVIRONMENT = False
+GLOBALLY_DELIVERABLE = True
+DEFAULT_TIMEOUT = 15  # secs
 
 # Based on RFC 2822 section 3.2.4 / RFC 5322 section 3.2.3, these
 # characters are permitted in email addresses (not taking into
@@ -20,8 +27,8 @@ DOT_ATOM_TEXT = '[' + ATEXT + ']+(?:\\.[' + ATEXT + ']+)*'
 # addresses to also include three specific ranges of UTF8 defined in
 # RFC3629 section 4, which appear to be the Unicode code points from
 # U+0080 to U+10FFFF.
-ATEXT_UTF8 = ATEXT + u"\u0080-\U0010FFFF"
-DOT_ATOM_TEXT_UTF8 = '[' + ATEXT_UTF8 + ']+(?:\\.[' + ATEXT_UTF8 + ']+)*'
+ATEXT_INTL = ATEXT + u"\u0080-\U0010FFFF"
+DOT_ATOM_TEXT_INTL = '[' + ATEXT_INTL + ']+(?:\\.[' + ATEXT_INTL + ']+)*'
 
 # The domain part of the email address, after IDNA (ASCII) encoding,
 # must also satisfy the requirements of RFC 952/RFC 1123 which restrict
@@ -36,6 +43,71 @@ EMAIL_MAX_LENGTH = 254
 LOCAL_PART_MAX_LENGTH = 64
 DOMAIN_MAX_LENGTH = 255
 
+# IANA Special Use Domain Names
+# Last Updated 2021-09-21
+# https://www.iana.org/assignments/special-use-domain-names/special-use-domain-names.txt
+#
+# The domain names without dots would be caught by the check that the domain
+# name in an email address must have a period, but this list will also catch
+# subdomains of these domains, which are also reserved.
+SPECIAL_USE_DOMAIN_NAMES = [
+    # The "arpa" entry here is consolidated from a lot of arpa subdomains
+    # for private address (i.e. non-routable IP addresses like 172.16.x.x)
+    # reverse mapping, plus some other subdomains. Although RFC 6761 says
+    # that application software should not treat these domains as special,
+    # they are private-use domains and so cannot have globally deliverable
+    # email addresses, which is an assumption of this library, and probably
+    # all of arpa is similarly special-use, so we reject it all.
+    "arpa",
+
+    # RFC 6761 says applications "SHOULD NOT" treat the "example" domains
+    # as special, i.e. applications should accept these domains.
+    #
+    # The domain "example" alone fails our syntax validation because it
+    # lacks a dot (we assume no one has an email address on a TLD directly).
+    # "@example.com/net/org" will currently fail DNS-based deliverability
+    # checks because IANA publishes a NULL MX for these domains, and
+    # "@mail.example[.com/net/org]" and other subdomains will fail DNS-
+    # based deliverability checks because IANA does not publish MX or A
+    # DNS records for these subdomains.
+    # "example", # i.e. "wwww.example"
+    # "example.com",
+    # "example.net",
+    # "example.org",
+
+    # RFC 6761 says that applications are permitted to treat this domain
+    # as special and that DNS should return an immediate negative response,
+    # so we also immediately reject this domain, which also follows the
+    # purpose of the domain.
+    "invalid",
+
+    # RFC 6762 says that applications "may" treat ".local" as special and
+    # that "name resolution APIs and libraries SHOULD recognize these names
+    # as special," and since ".local" has no global definition, we reject
+    # it, as we expect email addresses to be gloally routable.
+    "local",
+
+    # RFC 6761 says that applications (like this library) are permitted
+    # to treat "localhost" as special, and since it cannot have a globally
+    # deliverable email address, we reject it.
+    "localhost",
+
+    # RFC 7686 says "applications that do not implement the Tor protocol
+    # SHOULD generate an error upon the use of .onion and SHOULD NOT
+    # perform a DNS lookup.
+    "onion",
+
+    # Although RFC 6761 says that application software should not treat
+    # these domains as special, it also warns users that the address may
+    # resolve differently in different systems, and therefore it cannot
+    # have a globally routable email address, which is an assumption of
+    # this library, so we reject "@test" and "@*.test" addresses, unless
+    # the test_environment keyword argument is given, to allow their use
+    # in application-level test environments. These domains will generally
+    # fail deliverability checks because "test" is not an actual TLD.
+    "test",
+]
+
 # ease compatibility in type checking
 if sys.version_info >= (3,):
     unicode_class = str
@@ -47,8 +119,6 @@ else:
     DOT_ATOM_TEXT = DOT_ATOM_TEXT.decode("ascii")
     ATEXT_HOSTNAME = ATEXT_HOSTNAME.decode("ascii")
 
-DEFAULT_TIMEOUT = 15  # secs
-
 
 class EmailNotValidError(ValueError):
     """Parent class of all exceptions raised by this module."""
@@ -142,6 +212,8 @@ class ValidatedEmail(object):
 
     """Tests use this."""
     def __eq__(self, other):
+        if not isinstance(other, ValidatedEmail):
+            return False
         return (
             self.email == other.email
             and self.local_part == other.local_part
@@ -180,12 +252,26 @@ def __get_length_reason(addr, utf8=False, limit=EMAIL_MAX_LENGTH):
     return reason.format(prefix, diff, suffix)
 
 
+def caching_resolver(*, timeout=None, cache=None):
+    if timeout is None:
+        timeout = DEFAULT_TIMEOUT
+    resolver = dns.resolver.Resolver()
+    resolver.cache = cache or dns.resolver.LRUCache()
+    resolver.lifetime = timeout  # timeout, in seconds
+    return resolver
+
+
 def validate_email(
     email,
-    allow_smtputf8=True,
+    # /, # not supported in Python 3.6, 3.7
+    *,
+    allow_smtputf8=None,
     allow_empty_local=False,
-    check_deliverability=True,
-    timeout=DEFAULT_TIMEOUT,
+    check_deliverability=None,
+    test_environment=None,
+    globally_deliverable=GLOBALLY_DELIVERABLE,
+    timeout=None,
+    dns_resolver=None
 ):
     """
     Validates an email address, raising an EmailNotValidError if the address is not valid or returning a dict of
@@ -193,6 +279,16 @@ def validate_email(
     but if bytes it must be ASCII-only.
     """
 
+    # Fill in default values of arguments.
+    if allow_smtputf8 is None:
+        allow_smtputf8 = ALLOW_SMTPUTF8
+    if check_deliverability is None:
+        check_deliverability = CHECK_DELIVERABILITY
+    if test_environment is None:
+        test_environment = TEST_ENVIRONMENT
+    if timeout is None:
+        timeout = DEFAULT_TIMEOUT
+
     # Allow email to be a str or bytes instance. If bytes,
     # it must be ASCII because that's how the bytes work
     # on the wire with SMTP.
@@ -220,7 +316,7 @@ def validate_email(
     ret.smtputf8 = local_part_info["smtputf8"]
 
     # Validate the email address's domain part syntax and get a normalized form.
-    domain_part_info = validate_email_domain_part(parts[1])
+    domain_part_info = validate_email_domain_part(parts[1], test_environment=test_environment, globally_deliverable=globally_deliverable)
     ret.domain = domain_part_info["domain"]
     ret.ascii_domain = domain_part_info["ascii_domain"]
 
@@ -270,13 +366,14 @@ def validate_email(
             reason = "(when encoded in bytes)"
         raise EmailSyntaxError("The email address is too long {}.".format(reason))
 
-    if check_deliverability:
-        # Validate the email address's deliverability and update the
-        # return dict with metadata.
-        deliverability_info = validate_email_deliverability(ret["domain"], ret["domain_i18n"], timeout)
-        if "mx" in deliverability_info:
-            ret.mx = deliverability_info["mx"]
-            ret.mx_fallback_type = deliverability_info["mx-fallback"]
+    if check_deliverability and not test_environment:
+        # Validate the email address's deliverability using DNS
+        # and update the return dict with metadata.
+        deliverability_info = validate_email_deliverability(
+            ret["domain"], ret["domain_i18n"], timeout, dns_resolver
+        )
+        for key, value in deliverability_info.items():
+            setattr(ret, key, value)
 
     return ret
 
@@ -318,11 +415,11 @@ def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=Fals
 
     else:
         # The local part failed the ASCII check. Now try the extended internationalized requirements.
-        m = re.match(DOT_ATOM_TEXT_UTF8 + "\\Z", local)
+        m = re.match(DOT_ATOM_TEXT_INTL + "\\Z", local)
         if not m:
             # It's not a valid internationalized address either. Report which characters were not valid.
             bad_chars = ', '.join(sorted(set(
-                c for c in local if not re.match(u"[" + (ATEXT if not allow_smtputf8 else ATEXT_UTF8) + u"]", c)
+                unicodedata.name(c, repr(c)) for c in local if not re.match(u"[" + (ATEXT if not allow_smtputf8 else ATEXT_INTL) + u"]", c)
             )))
             raise EmailSyntaxError("The email address contains invalid characters before the @-sign: %s." % bad_chars)
 
@@ -336,6 +433,40 @@ def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=Fals
         # so we'll return the normalized local part in the return value.
         local = unicodedata.normalize("NFC", local)
 
+        # Check for unsafe characters.
+        # Some of this may be redundant with the range U+0080 to U+10FFFF that is checked
+        # by DOT_ATOM_TEXT_INTL.
+        for i, c in enumerate(local):
+            category = unicodedata.category(c)
+            if category[0] in ("L", "N", "P", "S"):
+                # letters, numbers, punctuation, and symbols are permitted
+                pass
+            elif category[0] == "M":
+                # combining character in first position would combine with something
+                # outside of the email address if concatenated to the right, but are
+                # otherwise permitted
+                if i == 0:
+                    raise EmailSyntaxError("The email address contains an initial invalid character (%s)."
+                                           % unicodedata.name(c, repr(c)))
+            elif category[0] in ("Z", "C"):
+                # spaces and line/paragraph characters (Z) and
+                # control, format, surrogate, private use, and unassigned code points (C)
+                raise EmailSyntaxError("The email address contains an invalid character (%s)."
+                                       % unicodedata.name(c, repr(c)))
+            else:
+                # All categories should be handled above, but in case there is something new
+                # in the future.
+                raise EmailSyntaxError("The email address contains a character (%s; category %s) that may not be safe."
+                                       % (unicodedata.name(c, repr(c)), category))
+
+        # Try encoding to UTF-8. Failure is possible with some characters like
+        # surrogate code points, but those are checked above. Still, we don't
+        # want to have an unhandled exception later.
+        try:
+            local.encode("utf8")
+        except ValueError:
+            raise EmailSyntaxError("The email address contains an invalid character.")
+
         # Flag that SMTPUTF8 will be required for deliverability.
         return {
             "local_part": local,
@@ -344,7 +475,7 @@ def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=Fals
         }
 
 
-def validate_email_domain_part(domain):
+def validate_email_domain_part(domain, test_environment=False, globally_deliverable=True):
     # Empty?
     if len(domain) == 0:
         raise EmailSyntaxError("There must be something after the @-sign.")
@@ -422,14 +553,31 @@ def validate_email_domain_part(domain):
     if not m:
         raise EmailSyntaxError("The email address contains invalid characters after the @-sign.")
 
-    # All publicly deliverable addresses have domain named with at least
-    # one period. We also know that all TLDs end with a letter.
-    if "." not in ascii_domain:
-        raise EmailSyntaxError("The domain name %s is not valid. It should have a period." % domain_i18n)
-    if not re.search(r"[A-Za-z]\Z", ascii_domain):
-        raise EmailSyntaxError(
-            "The domain name %s is not valid. It is not within a valid top-level domain." % domain_i18n
-        )
+    if globally_deliverable:
+        # All publicly deliverable addresses have domain named with at least
+        # one period, and we'll consider the lack of a period a syntax error
+        # since that will match people's sense of what an email address looks
+        # like. We'll skip this in test environments to allow '@test' email
+        # addresses.
+        if "." not in ascii_domain and not (ascii_domain == "test" and test_environment):
+            raise EmailSyntaxError("The domain name %s is not valid. It should have a period." % domain_i18n)
+
+        # We also know that all TLDs currently end with a letter.
+        if not re.search(r"[A-Za-z]\Z", ascii_domain):
+            raise EmailSyntaxError(
+                "The domain name %s is not valid. It is not within a valid top-level domain." % domain_i18n
+            )
+
+    # Check special-use and reserved domain names.
+    # Some might fail DNS-based deliverability checks, but that
+    # can be turned off, so we should fail them all sooner.
+    for d in SPECIAL_USE_DOMAIN_NAMES:
+        # See the note near the definition of SPECIAL_USE_DOMAIN_NAMES.
+        if d == "test" and test_environment:
+            continue
+
+        if ascii_domain == d or ascii_domain.endswith("." + d):
+            raise EmailSyntaxError("The domain name %s is a special-use or reserved name that cannot be used with email." % domain_i18n)
 
     # Return the IDNA ASCII-encoded form of the domain, which is how it
     # would be transmitted on the wire (except when used with SMTPUTF8
@@ -443,15 +591,26 @@ def validate_email_domain_part(domain):
     }
 
 
-def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT):
+def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT, dns_resolver=None):
     # Check that the domain resolves to an MX record. If there is no MX record,
     # try an A or AAAA record which is a deprecated fallback for deliverability.
+    # (Note that changing the DEFAULT_TIMEOUT module-level attribute
+    #  will not change the default value of this method's timeout argument.)
 
-    def dns_resolver_resolve_shim(resolver, domain, record):
+    # If no dns.resolver.Resolver was given, get dnspython's default resolver.
+    # Override the default resolver's timeout. This may affect other uses of
+    # dnspython in this process.
+    if dns_resolver is None:
+        dns_resolver = dns.resolver.get_default_resolver()
+        dns_resolver.lifetime = timeout
+
+    deliverability_info = {}
+
+    def dns_resolver_resolve_shim(domain, record):
         try:
             # dns.resolver.Resolver.resolve is new to dnspython 2.x.
             # https://dnspython.readthedocs.io/en/latest/resolver-class.html#dns.resolver.Resolver.resolve
-            return resolver.resolve(domain, record)
+            return dns_resolver.resolve(domain, record)
         except AttributeError:
             # dnspython 2.x is only available in Python 3.6 and later. For earlier versions
             # of Python, we maintain compatibility with dnspython 1.x which has a
@@ -460,7 +619,7 @@ def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT):
             # which we prevent by adding a "." to the domain name to make it absolute.
             # dns.resolver.Resolver.query is deprecated in dnspython version 2.x.
             # https://dnspython.readthedocs.io/en/latest/resolver-class.html#dns.resolver.Resolver.query
-            return resolver.query(domain + ".", record)
+            return dns_resolver.query(domain + ".", record)
 
     try:
         # We need a way to check how timeouts are handled in the tests. So we
@@ -469,36 +628,62 @@ def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT):
         if getattr(validate_email_deliverability, 'TEST_CHECK_TIMEOUT', False):
             raise dns.exception.Timeout()
 
-        resolver = dns.resolver.get_default_resolver()
-
-        if timeout:
-            resolver.lifetime = timeout
-
         try:
-            # Try resolving for MX records and get them in sorted priority order.
-            response = dns_resolver_resolve_shim(resolver, domain, "MX")
+            # Try resolving for MX records.
+            response = dns_resolver_resolve_shim(domain, "MX")
+
+            # For reporting, put them in priority order and remove the trailing dot in the qnames.
             mtas = sorted([(r.preference, str(r.exchange).rstrip('.')) for r in response])
-            mx_fallback = None
+
+            # Remove "null MX" records from the list (their value is (0, ".") but we've stripped
+            # trailing dots, so the 'exchange' is just ""). If there was only a null MX record,
+            # email is not deliverable.
+            mtas = [(preference, exchange) for preference, exchange in mtas
+                    if exchange != ""]
+            if len(mtas) == 0:
+                raise EmailUndeliverableError("The domain name %s does not accept email." % domain_i18n)
+
+            deliverability_info["mx"] = mtas
+            deliverability_info["mx_fallback_type"] = None
+
         except (dns.resolver.NoNameservers, dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
 
             # If there was no MX record, fall back to an A record.
             try:
-                response = dns_resolver_resolve_shim(resolver, domain, "A")
-                mtas = [(0, str(r)) for r in response]
-                mx_fallback = "A"
+                response = dns_resolver_resolve_shim(domain, "A")
+                deliverability_info["mx"] = [(0, str(r)) for r in response]
+                deliverability_info["mx_fallback_type"] = "A"
             except (dns.resolver.NoNameservers, dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
 
                 # If there was no A record, fall back to an AAAA record.
                 try:
-                    response = dns_resolver_resolve_shim(resolver, domain, "AAAA")
-                    mtas = [(0, str(r)) for r in response]
-                    mx_fallback = "AAAA"
+                    response = dns_resolver_resolve_shim(domain, "AAAA")
+                    deliverability_info["mx"] = [(0, str(r)) for r in response]
+                    deliverability_info["mx_fallback_type"] = "AAAA"
                 except (dns.resolver.NoNameservers, dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
 
                     # If there was no MX, A, or AAAA record, then mail to
                     # this domain is not deliverable.
                     raise EmailUndeliverableError("The domain name %s does not exist." % domain_i18n)
 
+        try:
+            # Check for a SPF reject all ("v=spf1 -all") record which indicates
+            # no emails are sent from this domain, which like a NULL MX record
+            # would indicate that the domain is not used for email.
+            response = dns_resolver_resolve_shim(domain, "TXT")
+            for rec in response:
+                value = b"".join(rec.strings)
+                if value.startswith(b"v=spf1 "):
+                    deliverability_info["spf"] = value.decode("ascii", errors='replace')
+                    if value == b"v=spf1 -all":
+                        raise EmailUndeliverableError("The domain name %s does not send email." % domain_i18n)
+        except dns.resolver.NoAnswer:
+            # No TXT records means there is no SPF policy, so we cannot take any action.
+            pass
+        except (dns.resolver.NoNameservers, dns.resolver.NXDOMAIN):
+            # Failure to resolve at this step will be ignored.
+            pass
+
     except dns.exception.Timeout:
         # A timeout could occur for various reasons, so don't treat it as a failure.
         return {
@@ -515,14 +700,10 @@ def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT):
             "There was an error while checking if the domain name in the email address is deliverable: " + str(e)
         )
 
-    return {
-        "mx": mtas,
-        "mx-fallback": mx_fallback,
-    }
+    return deliverability_info
 
 
 def main():
-    import sys
     import json
 
     def __utf8_input_shim(input_str):
@@ -536,10 +717,12 @@ def main():
         return output_str
 
     if len(sys.argv) == 1:
+        # Validate the email addresses pased line-by-line on STDIN.
+        dns_resolver = caching_resolver()
         for line in sys.stdin:
             email = __utf8_input_shim(line.strip())
             try:
-                validate_email(email)
+                validate_email(email, dns_resolver=dns_resolver)
             except EmailNotValidError as e:
                 print(__utf8_output_shim("{} {}".format(email, e)))
     else:
